/*** This Programs/Libraries are (C)opyright by S. Krahmer.
 *** You may use it under the terms of the GPL. You should have
 *** already received the file COPYING that shows you your rights. If not,
 *** you can get it at http://www.cs.uni-potsdam.de/homepages/students/linuxer
 *** the logit-package. You will also find some other nice utillities there.
 ***  
 *** THERE IS ABSOLUTELY NO WARRANTY. SO YOU USE IT AT YOUR OWN RISK.
 *** IT WAS WRITTEN IN THE HOPE THAT IT WILL BE USEFULL. I'M NOT RESPONSIBLE
 *** FOR ANY DAMAGE YOU MAYBE GET DUE TO USING MY PROGRAMS.
 ***/

/*! \file arp.cc
 */
#include <sys/types.h>
#include <string.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "usi++/usi-structs.h"
#include "usi++/usi++.h"

/*! Open device 'dev' for packet-capturing (ARP-packets)
 *  ARP-objects don't need to call init_device().
 */
ARP::ARP(char *dev, int promisc)
{
   	memset(&arphdr, 0, sizeof(arphdr));
	strcpy(filter_string, "arp");
        Datalink::init_device(dev, promisc, 100);
}

/*! \example arpw.cc
 */

ARP::~ARP()
{
}

/*! Return the source-hardware-adress of a ARP-packet
 */
char *ARP::get_sha(char *hwaddr, int len)
{
	char *s;
	memset(hwaddr, 0, len);
        
        /* switch over the hardware-layer of the ARP-packet */
	switch (ntohs(arphdr.ea_hdr.ar_hrd)) {
   	case ARPHRD_ETHER:
           	s = (char*)arphdr.arp_sha;
                snprintf(hwaddr, len, "%02x:%02x:%02x:%02x:%02x:%02x", (u_char)*s, 
                   	(u_char)*(s+1), (u_char)*(s+2), (u_char)*(s+3), 
                        (u_char)*(s+4), (u_char)*(s+5));
                break;
        default:
           	return NULL;
        }
        return hwaddr;
}

/*! Return the destination-hardware-adress.
 */
char *ARP::get_tha(char *hwaddr, int len)
{
	char *s;
	memset(hwaddr, 0, len);
        
        switch (ntohs(arphdr.ea_hdr.ar_hrd)) {
   	case ARPHRD_ETHER:
           	s = (char*)arphdr.arp_tha;
                snprintf(hwaddr, len, "%02x:%02x:%02x:%02x:%02x:%02x", (u_char)*s, 
                   	(u_char)*(s+1), (u_char)*(s+2), (u_char)*(s+3), 
                        (u_char)*(s+4), (u_char)*(s+5));
                break;
        default:
           	return NULL;
        }
        return hwaddr;
}

/*! Get target protocol-adress.
 *  Only IP is supportet yet!
 */
char *ARP::get_tpa(int r, char *pa, int len)
{
   	struct in_addr in;
        struct hostent *he;
        memset(pa, 0, len);
        
        // switch over protocol
	switch (ntohs(arphdr.ea_hdr.ar_pro)) {
	case ETH_P_IP:
                memcpy(&in, arphdr.arp_tpa, 4);
                if (!r || (he = gethostbyaddr((char*)&in, sizeof(in), AF_INET)) == NULL)
            	      	strncpy(pa, inet_ntoa(in), len);
                else
            	      	strncpy(pa, he->h_name, len);
                break;
        default:
           	return NULL;
        }
   	return pa;
}

/*! Get source protocol-adress.
 */
char *ARP::get_spa(int r, char *pa, int len)
{
   	struct in_addr in;
        struct hostent *he;
        memset(pa, 0, len);
        
	switch (ntohs(arphdr.ea_hdr.ar_pro)) {
	case ETH_P_IP:
                memcpy(&in, arphdr.arp_spa, 4);
                if (!r || (he = gethostbyaddr((char*)&in, sizeof(in), AF_INET)) == NULL)
            	      	strncpy(pa, inet_ntoa(in), len);
                else
            	      	strncpy(pa, he->h_name, len);
                break;
        default:
           	return NULL;
        }
   	return pa;
}

/*! Return the ARP-command.
 */
int ARP::get_op()
{
	return ntohs(arphdr.ea_hdr.ar_op);
}

/*! Sniff for an ARP-request/reply ...
 */
int ARP::sniffpack()
{
	return Datalink::sniffpack((char*)&arphdr, sizeof(arphdr));
}


