/*** This Programs/Libraries are (C)opyright by S. Krahmer.
 *** You may use it under the terms of the GPL. You should have
 *** already received the file COPYING that shows you your rights. If not,
 *** you can get it at http://www.cs.uni-potsdam.de/homepages/students/linuxer
 *** the logit-package. You will also find some other nice utillities there.
 ***  
 *** THERE IS ABSOLUTELY NO WARRANTY. SO YOU USE IT AT YOUR OWN RISK.
 *** IT WAS WRITTEN IN THE HOPE THAT IT WILL BE USEFULL. I AM NOT RESPONSIBLE
 *** FOR ANY DAMAGE YOU MAYBE GET DUE TO USING MY PROGRAMS.
 ***/

/*! \file datalink.cc
 *  \brief implementation of datalink-class
 */
#include <sys/types.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <netdb.h>
#include <sys/ioctl.h>
//#include <net/if.h>
#include "usi++/usi++.h"

Datalink::Datalink()
{  	
        /* Initialize */
        memset(&ether, 0, sizeof(ether));
        memset(host, 0, MAXHOSTLEN);
	memset(filter_string, 0, sizeof(filter_string));
        
        /* ask for local hostname */
        if (gethostname(host, MAXHOSTLEN) < 0) {
           	die("gethostname", PERROR, errno);
        }
	
	initialized = false;
        
}

/*! This constructor should be used to
 *  initialize raw-datalink-objects, means not IP/TCP/ICMP etc.
 */
Datalink::Datalink(char *filterStr)
{  	
        /* Initialize */
        memset(&ether, 0, sizeof(ether));
        memset(host, 0, MAXHOSTLEN);
	memset(filter_string, 0, sizeof(filter_string));
        strncpy(filter_string, filterStr, sizeof(filter_string));
        
        /* ask for local hostname */
        if (gethostname(host, MAXHOSTLEN) < 0) {
           	die("gethostname", PERROR, errno);
        }
	
	initialized = false;
        
}
        
Datalink::~Datalink()
{
}

int Datalink::get_datalink()
{
   	return datalink;
}

int Datalink::get_framelen()
{
	return framelen;
}

const char *Datalink::set_host(const char *st) 
{
   	strncpy(host, st, sizeof(host));
        return st;
}

/*! Fill buffer with src-hardware-adress of actuall packet,
 *  use 'datalink' to determine what HW the device is.
 *  Now only ethernet s supportet, but it's extensinable.
 */
char *Datalink::get_hwsrc(char *hwaddr, int len)
{
        unsigned char *s;
	memset(hwaddr, 0, len);
	
	switch (datalink) {
	case DLT_EN10MB:
		if (len < 2*ETH_ALEN)
           		return NULL;
    		s = ether.ether_shost;
    		sprintf(hwaddr, "%02x:%02x:%02x:%02x:%02x:%02x", (u_char)*s, 
	               (u_char)*(s+1), (u_char)*(s+2), (u_char)*(s+3), 
		       (u_char)*(s+4), (u_char)*(s+5));        
		break;
	default:
		return NULL;
	}
	return hwaddr;
}

/*! Fill buffer with dst-hardware-adress of actuall packet,
 *  use 'datalink' to determine what HW the device is.
 *  Now only ethernet s supportet, but it's extensinable.
 */
char *Datalink::get_hwdst(char *hwaddr, int len)
{
   	unsigned char *s;
	
	memset(hwaddr, 0, len);
        switch (datalink) {
        case DLT_EN10MB:
           	if (len < 2*ETH_ALEN)
           	       return NULL;
                s = ether.ether_dhost;
                sprintf(hwaddr, "%02x:%02x:%02x:%02x:%02x:%02x", (u_char)*s, 
	               (u_char)*(s+1), (u_char)*(s+2), (u_char)*(s+3), 
		       (u_char)*(s+4), (u_char)*(s+5));        
                break;
        default:
           	return NULL;
        }  	
	return hwaddr;
}

/*! Get protocol-type of ethernet-frame
 *  Maybe moves to ethernet-class in future?
 */
int Datalink::get_etype()
{
   	return ntohs(ether.ether_type);
}

/*! Initialize a device ("eth0" for example) for packet-
 *  capturing. It MUST be called before sniffpack() is launched.
 *  Set 'promisc' to 1 if you want the device running in promiscous mode.
 *  Fetch at most 'snaplen' bytes per call.
 */
int Datalink::init_device(char *dev, int promisc, int snaplen)
{
	char ebuf[PCAP_ERRBUF_SIZE];
	memset(ebuf, 0, PCAP_ERRBUF_SIZE);
   	
        if ((pd = pcap_open_live(dev, snaplen, promisc, 500, ebuf)) == NULL) {
           	die(ebuf, STDERR, 1);
        }
        
        if (pcap_lookupnet(dev, &localnet, &netmask, ebuf) < 0) {
           	snprintf(ebuf, sizeof(ebuf), "pcap_lookupnet: %s\n", pcap_geterr(pd));
		die(ebuf, STDERR, 1);
        }
        

	/* The filter_string must be filled by derived classes, such
	 * as IP, where the virtual init_device() simply sets filter_string
	 * to "ip" and then calls Datalink::init_device().
	 */
        if (pcap_compile(pd, &filter, filter_string, 1, netmask) < 0) {
           	snprintf(ebuf, sizeof(ebuf), "pcap_compile: %s\n", pcap_geterr(pd));
		die(ebuf, STDERR, 1);
        }
        if (pcap_setfilter(pd, &filter) < 0) {
           	snprintf(ebuf, sizeof(ebuf), "pcap_setfilter: %s\n", pcap_geterr(pd));
		die(ebuf, STDERR, 1);
        }
        if ((datalink = pcap_datalink(pd)) < 0) {
           	snprintf(ebuf, sizeof(ebuf), "pcap_datalink: %s\n", pcap_geterr(pd));
                die(ebuf, STDERR, 1);
        }
        
        	
        // turn datalink into framelen
        switch (datalink) {
        case DLT_EN10MB:
		framelen = sizeof(ether);
                break;
        case DLT_PPP:
		framelen = 4;
                break;
	case DLT_PPP_BSDOS:
		framelen = 24;
		break;
        case DLT_SLIP:
		framelen = SLIP_LEN;
                break;
	case DLT_RAW:
           	framelen = 0;
		break;
        default:
           	die("Datalink::init_device: Unknown datalink.\n", STDERR, 1);
        }
        
	initialized = true;
        return 0;
}


/*! set a new filter for capturing
 */
int Datalink::setfilter(char *s)
{
	char ebuf[PCAP_ERRBUF_SIZE];
	memset(ebuf, 0, PCAP_ERRBUF_SIZE);
	
	if (!initialized) 
		die("(Datalink::setfilter) Device not initialized.\n", STDERR, 1);

   	memset(filter_string, 0, 1000);
	snprintf(filter_string, 1000, "%s", s);
        
        if (pcap_compile(pd, &filter, filter_string, 1, netmask) < 0) {
           	snprintf(ebuf, sizeof(ebuf), "pcap_compile: %s\n", pcap_geterr(pd));
		die(ebuf, STDERR, 1);
        }

        if (pcap_setfilter(pd, &filter) < 0) {
           	snprintf(ebuf, sizeof(ebuf), "pcap_setfilter: %s\n", pcap_geterr(pd));
		die(ebuf, STDERR, 1);
        }
        return 0;
}

int Datalink::sniffpack(char *s, int len)
{
   	char *tmp;        
        
	memset(s, 0, len);
	
	if (!initialized)
		die("(Datalink::sniffpack) Device not initialized.\n", STDERR, 1);
                
        while ((tmp = (char*)pcap_next(pd, &phdr)) == NULL);

        switch (datalink) {
        case DLT_EN10MB:
                memcpy(&ether, tmp, framelen);
                break;
        case DLT_PPP:
                break;
	case DLT_PPP_BSDOS:
		break;
        case DLT_SLIP:
                break;
	case DLT_RAW:
		break;
        default:
           	die("Datalink::sniffpack: Unknown datalink.\n", STDERR, 1);
        }

        
	/* framelen was already calculated by init_device */
        memcpy(s, (tmp + framelen), 
	       phdr.len - framelen < len ? phdr.len - framelen : len);
	       
        return (phdr.len - framelen);
}     
  
#if 0
/*! Return the MAC-adress of device 'device'.
 */
char *getMAC(const char *dev, char *macstring, int len)
{
        struct ifreq ifr;
        int sfd;

	memset(&ifr, 0, sizeof(ifr));
	memset(macstring, 0, len);
	
        if ((sfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
           	die("socket", PERROR, errno);
        }

        snprintf(ifr.ifr_name, sizeof(ifr.ifr_name), "%s", dev);
        if (ioctl(sfd, SIOCGIFHWADDR, &ifr) < 0) {
           	die("ioctl", PERROR, errno);
        }
        close(sfd);
        char *s = ifr.ifr_hwaddr.sa_data;
        snprintf(macstring, len, "%02x:%02x:%02x:%02x:%02x:%02x", (u_char)*s, 
	                          (u_char)*(s+1), (u_char)*(s+2), (u_char)*(s+3), 
			          (u_char)*(s+4), (u_char)*(s+5));
        return macstring;
}

/*! Set MAC of device 'device'.
 *  'mac' is something like "11:22:ff:77:aa:11"
 *  XXX: Doesn't work yet!
 */
char *setMAC(const char *dev, char *mac)
{
        struct ifreq ifr;
        int sfd;
	unsigned char s[6];

	memset(&ifr, 0, sizeof(ifr));
        if ((sfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
           	die("socket", PERROR, errno);
        }

        snprintf(ifr.ifr_name, sizeof(ifr.ifr_name), "%s", dev);
	sscanf(mac, "%02x:%02x:%02x:%02x:%02x:%02x", &s[0], &s[1], &s[2],
	            &s[3], &s[4], &s[5]);
        /*for (int i = 0; i < 6; i++) 
		printf("%02x\n", s[i]);*/
        
	memcpy(ifr.ifr_hwaddr.sa_data, s, 6);
        
        if (ioctl(sfd, SIOCSIFHWADDR, &ifr) < 0) {
           	die("ioctl", PERROR, errno);
        }
        close(sfd);
        return mac;
}
#endif

