/*** This Programs/Libraries are (C)opyright by S. Krahmer.
 *** You may use it under the terms of the GPL. You should have
 *** already received the file COPYING that shows you your rights. If not,
 *** you can get it at http://www.cs.uni-potsdam.de/homepages/students/linuxer
 *** the logit-package. You will also find some other nice utillities there.
 ***  
 *** THERE IS ABSOLUTELY NO WARRANTY. SO YOU USE IT AT YOUR OWN RISK.
 *** IT WAS WRITTEN IN THE HOPE THAT IT WILL BE USEFULL. I AM NOT RESPONSIBLE
 *** FOR ANY DAMAGE YOU MAYBE GET DUE TO USING MY PROGRAMS.
 ***/
 
/*! \file icmp.cc
 *  \brief Implementation of ICMP-class.
 */ 
#include <stdio.h>
#include "config.h"
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <iostream.h>
#include "usi++/usi++.h"

ICMP::ICMP(const char* host) 
#ifndef IPPROTO_ICMP
#define IPPROTO_ICMP 1
#endif
      : IP(host, IPPROTO_ICMP)
{        
        // clear memory
        memset(&icmphdr, 0, sizeof(icmphdr));
                
}

/*! Set the type-field in the actuall ICMP-packet.
 */
int ICMP::set_type(u_int8_t t)
{       
        icmphdr.type = t;
        return 0;
}

/*! Get the type-field from the actuall ICMP-packet.
 */
/*! \example hardSpy.cc
 */
u_int8_t ICMP::get_type()
{
   	return icmphdr.type;
}

int ICMP::set_code(u_int8_t c)
{
   	icmphdr.code = c;
        return 0;
}

u_int8_t ICMP::get_code()
{
   	return icmphdr.code;
}

int ICMP::set_gateway(u_int32_t g)
{
   	icmphdr.un.gateway = htonl(g);
        return 0;
}

u_int32_t ICMP::get_gateway()
{
   	return ntohl(icmphdr.un.gateway);
}

int ICMP::set_mtu(u_int16_t mtu)
{
   	icmphdr.un.frag.mtu = mtu;
   	return 0;
}

u_int16_t ICMP::get_mtu()
{
   	return icmphdr.un.frag.mtu;
}

/*! Set id field in the actuall ICMP-packet 
 */
int ICMP::set_icmpId(u_int16_t id)
{
   	icmphdr.un.echo.id = id;
        return 0;
}

/*! Get the id field from actuall ICMP-packet.
 */
u_int16_t ICMP::get_icmpId()
{
   	return icmphdr.un.echo.id;
}

/*! Set the sequecenumber of the actuall ICMP-packet.
 */
int ICMP::set_seq(u_int16_t s)
{
   	icmphdr.un.echo.sequence = s;
        return 0;
}

/*! Get the sequence-number of actuall ICMP-packet
 */
u_int16_t ICMP::get_seq()
{
   	return icmphdr.un.echo.sequence;
}

/*! send an ICMP-packet containing 'payload' which
 *  is 'paylen' bytes long
 */
int ICMP::sendpack(void *payload, int paylen)
{
   	int len = sizeof(struct usipp::icmphdr) + paylen;	// the packetlenght
                
        struct usipp::icmphdr *i;

        // s will be our packet
   	char *s = new char[len];
        memset(s, 0, len);
        
        // copy ICMP header to packet
        memcpy((char*)s, (struct usipp::icmphdr*)&this->icmphdr, sizeof(icmphdr));

   	if (payload)
           	memcpy(s+sizeof(icmphdr), payload, paylen);

        i = (struct usipp::icmphdr*)s;
        
        // calc checksum over packet
        //i->sum = 0;
	
	if (i->sum == 0)
		i->sum = in_cksum((unsigned short*)s, len);
        
        return IP::sendpack(s, len);
        
}

/*! send a ICMP-packet with string 'payload' as payload.
 */
int ICMP::sendpack(char *payload)
{
   	return sendpack(payload, strlen(payload));
}

/*! receive an ICMP-packet, not sniff it. This means
 *  we will only get packets with a destinationadress that
 *  belongs to one of our devices.
 */
int ICMP::recvpack(void *s, int len)
{

	int plen = len + sizeof(struct usipp::icmphdr);
   	char *tmp = new char[plen];
        
   	int r = IP::recvpack(tmp, plen);
	
        // point to ICMP header
        struct usipp::icmphdr *icmph = (struct usipp::icmphdr*)(tmp);
        
        // save ICMP header for public functions
        memcpy(&icmphdr, icmph, sizeof(struct usipp::icmphdr));
        
        // and give user the payload
        if (s)
		memcpy(s, ++icmph, len);
        
        delete[] tmp;
        return r - sizeof(struct usipp::icmphdr);
}    

/*! handle packets, that are NOT actually for the
 *  local adress!
 */
int ICMP::sniffpack(void *s, int len)
{
	int plen = len + sizeof(struct usipp::icmphdr);
   	char *tmp = new char[plen];
        int r = 0;
	memset(s, 0, len);
	memset(tmp, 0, plen);
	
   	r = IP::sniffpack(tmp, plen);
	
        // point to ICMP header
        struct usipp::icmphdr *icmph = (struct usipp::icmphdr*)(tmp);
        
        // save ICMP header for public functions
        memcpy(&icmphdr, icmph, sizeof(struct usipp::icmphdr));
        
        // and give user the payload
	if (s)
    		memcpy(s, ++icmph, len);
        
        delete[] tmp;
        return r - sizeof(struct usipp::icmphdr);
}    

/*! Initialize a device ("eth0" for example) for packet-
 *  capturing. It MUST be called before sniffpack() is launched.
 *  Set 'promisc' to 1 if you want the device running in promiscous mode.
 *  Fetch at most 'snaplen' bytes per call.
 */
int ICMP::init_device(char *dev, int promisc, int snaplen)
{
   	strcpy(filter_string, "icmp");
        Datalink::init_device(dev, promisc, snaplen);
        return 0;
}

