/*** This Programs/Libraries are (C)opyright by S. Krahmer.
 *** You may use it under the terms of the GPL. You should have
 *** already received the file COPYING that shows you your rights. If not,
 *** you can get it at http://www.cs.uni-potsdam.de/homepages/students/linuxer
 *** the logit-package. You will also find some other nice utillities there.
 ***  
 *** THERE IS ABSOLUTELY NO WARRANTY. SO YOU USE IT AT YOUR OWN RISK.
 *** IT WAS WRITTEN IN THE HOPE THAT IT WILL BE USEFULL. I AM NOT RESPONSIBLE
 *** FOR ANY DAMAGE YOU MAYBE GET DUE TO USING MY PROGRAMS.
 ***/

/*! \file tcp.cc 
 *  \brief Implementation of TCP-class.
 */
#include <string.h>
#include <stdlib.h>
#include "config.h"
#include <errno.h>
#include "usi++/usi++.h"


TCP::TCP(const char *host)
#ifndef IPPROTO_TCP
#define IPPROTO_TCP 6
#endif
     : IP(host, IPPROTO_TCP)
{
	srand(time(NULL));
   	memset(&tcph, 0, sizeof(tcph));
        memset(&pseudo, 0, sizeof(pseudo));
	memset(tcpOptions, 0, sizeof(tcpOptions));
	
        tcph.th_off = 5;
        tcph.th_ack = htonl(rand());
        tcph.th_seq = htonl(rand());
}

/*! Get the sourceport in human-readable form.
 */
u_int16_t TCP::get_srcport()
{
   	return ntohs(tcph.th_sport);
}

/*! Get the destinationport in human-readable form.
 */
u_int16_t TCP:: get_dstport()
{
   	return ntohs(tcph.th_dport);
}

/*! Get TCP-sequencenumber
 */
u_int32_t TCP::get_seq()
{
   	return ntohl(tcph.th_seq);
}

/*! Get the actual achnkowledge-number from the TCP-header.
 */
u_int32_t TCP::get_ack()
{
   	return ntohl(tcph.th_ack);
}

u_int8_t TCP::get_off()
{
   	return tcph.th_off;
}

/*! Set TCP-flags 
 */
u_int8_t TCP::get_flags()
{
   	return tcph.th_flags;
}

u_int16_t TCP::get_win()
{
   	return ntohs(tcph.th_win);
}

/*! Get TCP-header checksum
 */
u_int16_t TCP::get_tcpsum()
{
   	return tcph.th_sum;
}

u_int16_t TCP::get_urg()
{
   	return tcph.th_urp;
}

int TCP::set_srcport(u_int16_t sp)
{
   	tcph.th_sport = htons(sp);
        return 0;
}

int TCP::set_dstport(u_int16_t dp)
{
   	tcph.th_dport = htons(dp);
        return 0;
}

/*! Set the sequencenumber-filed in the TCP-header.
 */
int TCP::set_seq(u_int32_t s)
{
   	tcph.th_seq = htonl(s);
        return 0;
}

/*! Set the acknowledgenumber-filed in the TCP-header.
 *  This is only monitored by the target-kernel, if TH_ACK
 *  is set in the TCP-flags.
 */
int TCP::set_ack(u_int32_t a)
{
   	tcph.th_ack = htonl(a);
        return 0;
}

int TCP::set_off(u_int8_t o)
{
   	tcph.th_off = o;
        return 0;
}

/*! Set TCP-flags
 */
int TCP::set_flags(u_int8_t f)
{
   	tcph.th_flags = f;
        return 0;
}

int TCP::set_win(u_int16_t w)
{
   	tcph.th_win = htons(w);
        return 0;
}

/*! Set TCP-checksum. Calling this function with s != 0
 *  will prevent sendpack from calculating the checksum!!!
 */
int TCP::set_tcpsum(u_int16_t s)
{
   	tcph.th_sum = s;
        return 0;
}

int TCP::set_urg(u_int16_t u)
{
   	tcph.th_urp = u;
        return 0;
}

/*! Send a TCP-packet
 */
int TCP::sendpack(void *buf, int paylen)
{
	int len = paylen + sizeof(tcph) + sizeof(pseudo);
	char *tmp = new char[len];
	memset(tmp, 0, len);

   	// build a pseudoheader for IP-checksum, as
        // required per RFC 793
	pseudo.saddr = get_src();	// sourceadress
	pseudo.daddr = get_dst();	// destinationadress
	pseudo.zero = 0;
	pseudo.proto = IPPROTO_TCP;
	pseudo.len = htons(sizeof(tcph) + paylen);
	
	//tcph.th_sum = 0;
	
        // copy pseudohdr+header+data to buffer
	memcpy(tmp, &pseudo, sizeof(pseudo));
	memcpy(tmp + sizeof(pseudo), &tcph, sizeof(tcph));
	memcpy(tmp + sizeof(pseudo) + sizeof(tcph), buf, paylen);
	
        // calc checksum over it
	struct usipp::tcphdr *t = (struct usipp::tcphdr*)(tmp + sizeof(pseudo));
	
	if (tcph.th_sum == 0)
		t->th_sum = in_cksum((unsigned short*)tmp, len);
	
	IP::sendpack(tmp + sizeof(pseudo), len - sizeof(pseudo));
	
	delete [] tmp;
	return 0;
}


int TCP::sendpack(char *s)
{
	return sendpack(s, strlen(s));
}

/*! Receive a TCP-packet, destinated for this host.
 */
int TCP::recvpack(void *buf, int len)
{  	
	int xlen = len + sizeof(tcph) + sizeof(tcpOptions);
	
        char *tmp = new char[xlen];
        memset(tmp, 0, xlen);
        memset(buf, 0, len);
	
        int r = IP::recvpack(tmp, xlen);
	int tcplen = tcph.th_off<<2;
	
	if (tcplen > 60)
		tcplen = 60;
	
        memcpy(&tcph, tmp, tcplen);
        
	if (buf)
		memcpy(buf, tmp + tcplen, len);
        
        delete [] tmp;
        return r - tcplen;
}

/*! Sniff a TCP-packet.
 */
int TCP::sniffpack(void *buf, int len)
{  	
        int xlen = len + sizeof(tcph) + sizeof(tcpOptions);
	
	char *tmp = new char[xlen];
	int r = 0;
	
        memset(tmp, 0, xlen);
        memset(buf, 0, len);
	
        r = IP::sniffpack(tmp, xlen);
	int tcplen = tcph.th_off<<2;
	
	if (tcplen > 60)
		tcplen = 60;
	
        memcpy(&tcph, tmp, tcplen);
        
	if (buf)
		memcpy(buf, tmp + tcplen, len);
        
        delete [] tmp;
        return r - tcplen;
}
        
 
/*! Initialize a device ("eth0" for example) for packet-
 *  capturing. It MUST be called before sniffpack() is launched.
 *  Set 'promisc' to 1 if you want the device running in promiscous mode.
 *  Fetch at most 'snaplen' bytes per call.
 */
int TCP::init_device(char *dev, int promisc, int snaplen)
{
   	strcpy(filter_string, "tcp");
        Datalink::init_device(dev, promisc, snaplen);
        return 0;
}

