/*** This Programs/Libraries are (C)opyright by S. Krahmer.
 *** You may use it under the terms of the GPL. You should have
 *** already received the file COPYING that shows you your rights. If not,
 *** you can get it at http://www.cs.uni-potsdam.de/homepages/students/linuxer
 *** the logit-package. You will also find some other nice utillities there.
 ***  
 *** THERE IS ABSOLUTELY NO WARRANTY. SO YOU USE IT AT YOUR OWN RISK.
 *** IT WAS WRITTEN IN THE HOPE THAT IT WILL BE USEFULL. I AM NOT RESPONSIBLE
 *** FOR ANY DAMAGE YOU MAYBE GET DUE TO USING MY PROGRAMS.
 ***/
 
/*! \file udp.cc
 *  \brief Implementation of UDP-class.
 */ 
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "usi++/usi++.h"

UDP::UDP(const char *host)
#ifndef IPPROTO_UDP
#define IPPROTO_UDP 17
#endif
      : IP(host, IPPROTO_UDP)
{
   	memset(&udph, 0, sizeof(udph));        
        memset(&pseudo, 0, sizeof(pseudo));
}

/*! Get the sourceport of UDP-datagram.
 */
u_int16_t UDP::get_srcport()
{
   	return ntohs(udph.source);
}

/*! Get the destinationport of the UDP-datagram
 */
u_int16_t UDP::get_dstport()
{
   	return ntohs(udph.dest);
}

/*! Return length of UDP-header plus contained data.
 */
u_int16_t UDP::get_len()
{
   	return ntohs(udph.len);
}

/*! Return the checksum of UDP-datagram.
 */
u_int16_t UDP::get_udpsum()
{
   	return udph.check;
}

/*! Set the sourceport in the UDP-header.
 */
int UDP::set_srcport(u_int16_t sp)
{
   	udph.source = htons(sp);
        return 0;
}

/*! Set the destinationport in the UDP-header.
 */
int UDP::set_dstport(u_int16_t dp)
{
   	udph.dest = htons(dp);
        return 0;
}

/*! Set the length of the UDP-datagramm.
 */
int UDP::set_len(u_int16_t l)
{
   	udph.len = htons(l);
        return 0;
}

/*! Set the UDP-checksum. Calling this function with s != 0
 *  will prevent sendpack() from setting the checksum!!!
 */
int UDP::set_udpsum(u_int16_t s)
{
   	udph.check = s;
        return 0;
}

/*! Send an UDP-datagramm, containing 'paylen' bytes of data.
 */
int UDP::sendpack(void *buf, int paylen)
{
	int len = paylen + sizeof(udph) + sizeof(pseudo);
	char *tmp = new char[len];
	memset(tmp, 0, len);

   	// build a pseudoheader for IP-checksum, as
        // required per RFC ???	
	pseudo.saddr = get_src();	// sourceadress
	pseudo.daddr = get_dst();	// destinationadress
	pseudo.zero = 0;
	pseudo.proto = IPPROTO_UDP;
	pseudo.len = htons(sizeof(udph) + paylen);
	
	udph.len = htons(paylen + sizeof(udph));
	//udph.check = 0;
	
        // copy pseudohdr+header+data to buffer
	memcpy(tmp, &pseudo, sizeof(pseudo));
	memcpy(tmp + sizeof(pseudo), &udph, sizeof(udph));
	memcpy(tmp + sizeof(pseudo) + sizeof(udph), buf, paylen);
	
        // calc checksum over it
	struct usipp::udphdr *u = (struct usipp::udphdr*)(tmp + sizeof(pseudo));
	
	if (udph.check == 0)
		u->check = in_cksum((unsigned short*)tmp, len);
	
	IP::sendpack(tmp + sizeof(pseudo), len - sizeof(pseudo));
	
	delete [] tmp;
	return 0;
}


int UDP::sendpack(char *s)
{
	return sendpack(s, strlen(s));
}

/*! Receive an UDP-datagramm.
 */
int UDP::recvpack(void *buf, int len)
{  	
        char *tmp = new char[len+sizeof(udph)];
        memset(tmp, 0, len + sizeof(udph));
	memset(buf, 0, len);
        
        int r = IP::recvpack(tmp, len + sizeof(udph));
        memcpy(&udph, tmp, sizeof(udph));
        
	if (buf)
		memcpy(buf, tmp + sizeof(udph), len);
        
        delete [] tmp;
        return r - sizeof(udph);
}
        
/*! Capture packets that are not for our host.
 */ 
int UDP::sniffpack(void *buf, int len)
{  	
        char *tmp = new char[len+sizeof(udph)];
	int r = 0;
        memset(tmp, 0, len + sizeof(udph));
        
        r = IP::sniffpack(tmp, len + sizeof(udph));
        memcpy(&udph, tmp, sizeof(udph));

	if (buf)
    		memcpy(buf, tmp + sizeof(udph), len);
        
        delete [] tmp;
        return r-sizeof(udph);
}

/*! Initialize a device ("eth0" for example) for packet-
 *  capturing. It MUST be called before sniffpack() is launched.
 *  Set 'promisc' to 1 if you want the device running in promiscous mode.
 *  Fetch at most 'snaplen' bytes per call.
 */
int UDP::init_device(char *dev, int promisc, int snaplen)
{
   	strcpy(filter_string, "udp");
        Datalink::init_device(dev, promisc, snaplen);
        return 0;
}

